const SimpleDB = require('../dynamodb');
const expect = require('chai').expect;
const AWS = require('aws-sdk');

const config = {
    region: 'local',
    endpoint: 'http://127.0.0.1:8000',
    accessKeyId: 'not-important',
    secretAccessKey: 'no-important'
}

let params = {
    TableName: 'MyTable',
    KeySchema: [
        {
            AttributeName: 'name',
            KeyType: 'HASH'
        }
    ],
    AttributeDefinitions: [
        {
            AttributeName: 'name',
            AttributeType: 'S'
        }
    ],
    ProvisionedThroughput:  {
        ReadCapacityUnits: 1,
        WriteCapacityUnits: 1
    }
}

function createTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.createTable(params, function (err, data) {
            if (err) {
                reject(err);
                return;
            }
 
            resolve(data);
        });
    })
}

function deleteTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.deleteTable({ TableName: 'MyTable' }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve(data);
        })
    })
}

before(() => {
    return createTable();
});

after(() => {
    return deleteTable();
});

describe('1 DynamoDB Put Operation', () => {
    describe('1.1 Param Building Test', () => {
        it('1.1.1 Return All New', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.put()
                .tableName('MyTable')
                .item({ name: 'Alice' })
                .attribute('age').equal(20).and().attribute('job').equal('retired')
                .returnValues().allNew()
                .returnItemCollectionMetrics()
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('MyTable');
                    expect(params.Item).to.deep.equal({ name: 'Alice' });
                    expect(params.ConditionExpression).to.be.equal('age = :age AND job = :job');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':age': 20, ':job': 'retired' });
                    expect(params.ReturnItemCollectionMetrics).to.be.equal('SIZE');
                    expect(params.ReturnValues).to.be.equal('ALL_NEW');
                });
        });
    
        it('1.1.2 Return All Old', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.put()
                .tableName('MyTable')
                .item({ name: 'Alice' })
                .attribute('age').equal(20).and().attribute('job').equal('retired')
                .returnValues().allOld()
                .returnItemCollectionMetrics()
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('MyTable');
                    expect(params.Item).to.deep.equal({ name: 'Alice' });
                    expect(params.ConditionExpression).to.be.equal('age = :age AND job = :job');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':age': 20, ':job': 'retired' });
                    expect(params.ReturnItemCollectionMetrics).to.be.equal('SIZE');
                    expect(params.ReturnValues).to.be.equal('ALL_OLD');
                });
        });
    
        it('1.1.3 Return Updated New', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.put()
                .tableName('MyTable')
                .item({ name: 'Alice' })
                .attribute('age').equal(20).and().attribute('job').equal('retired')
                .returnValues().updatedNew()
                .returnItemCollectionMetrics()
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('MyTable');
                    expect(params.Item).to.deep.equal({ name: 'Alice' });
                    expect(params.ConditionExpression).to.be.equal('age = :age AND job = :job');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':age': 20, ':job': 'retired' });
                    expect(params.ReturnItemCollectionMetrics).to.be.equal('SIZE');
                    expect(params.ReturnValues).to.be.equal('UPDATED_NEW');
                });
        });
    
        it('1.1.4 Return Updated Old', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.put()
                .tableName('MyTable')
                .item({ name: 'Alice' })
                .attribute('age').equal(20).and().attribute('job').equal('retired')
                .returnValues().updatedOld()
                .returnItemCollectionMetrics()
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('MyTable');
                    expect(params.Item).to.deep.equal({ name: 'Alice' });
                    expect(params.ConditionExpression).to.be.equal('age = :age AND job = :job');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':age': 20, ':job': 'retired' });
                    expect(params.ReturnItemCollectionMetrics).to.be.equal('SIZE');
                    expect(params.ReturnValues).to.be.equal('UPDATED_OLD');
                });
        });    
    });

    describe('1.2 Query Test', () => {
        it('1.2.1 Put Item', () => {
            let db = new SimpleDB(config);
            return db.put()
                .tableName('MyTable')
                .item({
                    name: 'Alice',
                    age: 20,
                    job: 'retired'
                })
                .run();
        });
    })
});    

