const SimpleDB = require('../dynamodb');
const expect = require('chai').expect;
const AWS = require('aws-sdk');

const config = {
    region: 'local',
    endpoint: 'http://127.0.0.1:8000',
    accessKeyId: 'not-important',
    secretAccessKey: 'no-important'
}

const TABLE_NAME ='MyDeleteTestTable'

let params = {
    TableName: TABLE_NAME,
    KeySchema: [
        {
            AttributeName: 'userId',
            KeyType: 'HASH'
        },
        {
            AttributeName: 'timestamp',
            KeyType: 'RANGE'
        },
    ],
    AttributeDefinitions: [
        {
            AttributeName: 'userId',
            AttributeType: 'S'
        },
        {
            AttributeName: 'timestamp',
            AttributeType: 'S'
        }
    ],
    ProvisionedThroughput:  {
        ReadCapacityUnits: 1,
        WriteCapacityUnits: 1
    }
}

let timestampValue = new Date().toISOString();


function createTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.createTable(params, function (err, data) {
            if (err) {
                reject(err);
                return;
            }
 
            resolve(data);
        });
    })
}

function deleteTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.deleteTable({ TableName: TABLE_NAME }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve(data);
        })
    })
}

function prepopulateTable() {
    AWS.config.update(config);
    let dynamodb = new AWS.DynamoDB();
    let docClient = new AWS.DynamoDB.DocumentClient({
        options: dynamodb,
    });
    return new Promise((resolve, reject) => {
        docClient.put({
            TableName: TABLE_NAME,
            Item: {
                userId: '123',
                timestamp: timestampValue
            }
        }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve(data);
        })
    })
}

before(() => {
    return createTable().then(() => {
        return prepopulateTable();
    })
});

after(() => {
    return deleteTable();
});

describe('3 DynamoDB Delete Operation', () => {
    describe('3.1 Param Test', () => {
        it('3.1.1 Basic', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.delete()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .sortKey('timestamp').equal('2011-12-19T15:28:46.493Z')
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal(TABLE_NAME);
                    expect(params.Key).to.deep.equal({ userId: '123', timestamp: '2011-12-19T15:28:46.493Z' });
                });
        });
    });

    describe('3.2 Query Test', () => {
        it('3.2.1 Basic Delete', () => {
            let db = new SimpleDB(config);
            return db.delete()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .sortKey('timestamp').equal(timestampValue)
                .run();
        });

        it('3.2.2 Check Deletion', () => {
            let db = new SimpleDB(config);
            return db.get()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .sortKey('timestamp').equal(timestampValue)
                .run()
                .then(result => {
                    expect(result).to.be.empty;
                    // console.log(result);
                });
        });
    })
    
});