const SimpleDB = require('../dynamodb');
const chai = require('chai');
const chaiAsPromised = require('chai-as-promised');
chai.use(chaiAsPromised);
const expect = chai.expect;
chai.should();

const AWS = require('aws-sdk');

const config = {
    region: 'local',
    endpoint: 'http://127.0.0.1:8000',
    accessKeyId: 'not-important',
    secretAccessKey: 'no-important'
}

const TABLE_NAME ='MyUpdateTestTable'

let params = {
    TableName: TABLE_NAME,
    KeySchema: [
        {
            AttributeName: 'userId',
            KeyType: 'HASH'
        },

    ],
    AttributeDefinitions: [
        {
            AttributeName: 'userId',
            AttributeType: 'S'
        },

    ],
    ProvisionedThroughput:  {
        ReadCapacityUnits: 1,
        WriteCapacityUnits: 1
    }
}

let timestampValue = new Date().toISOString();


function createTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.createTable(params, function (err, data) {
            if (err) {
                reject(err);
                return;
            }
 
            resolve(data);
        });
    })
}

function deleteTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.deleteTable({ TableName: TABLE_NAME }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve(data);
        })
    })
}

function prepopulateTable() {
    AWS.config.update(config);
    let dynamodb = new AWS.DynamoDB();
    let docClient = new AWS.DynamoDB.DocumentClient({
        options: dynamodb,
    });
    return new Promise((resolve, reject) => {
        docClient.put({
            TableName: TABLE_NAME,
            Item: {
                userId: '123',
                timestamp: timestampValue,
                phone: {
                    home: [ '87654321' ]
                }
            }
        }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve(data);
        })
    })
}

before(() => {
    return createTable().then(() => {
        return prepopulateTable();
    })
        .catch(err => {
            return Promise.resolve();
        });
});

after(() => {
    return deleteTable();
});

describe('4 DynamoDB Update Operation', () => {
    describe('4.1 Param Test', () => {
        it('4.1.1 SET', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .sortKey('name').equal('alice')
                .set().attribute('age').equal(20)
                .set().attribute('job').equal('retired')
                .where().attribute('height').lessThan(150)
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal(TABLE_NAME);
                    expect(params.Key).to.deep.equal({ userId: '123', name: 'alice' });
                    expect(params.UpdateExpression).to.be.equal('SET age = :age, job = :job');
                    expect(params.ConditionExpression).to.be.equal('height < :height');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':height': 150, ':age': 20, ':job': 'retired' });
                });
        });
    
        it('4.1.2 SET nested map attributes', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .set().attribute('pr.5star[1]').equal(20)
                .set().attribute('PR.3star').equal('retired')
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('SET pr.5star[1] = :pr_5star_1_, PR.3star = :PR_3star');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':pr_5star_1_': 20, ':PR_3star': 'retired' });
                });
        });
    
        it('4.1.3 SET append element to list', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .set().attribute('ri').append(['Hacksaw', 'Screwdriver'])
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('SET ri = list_append(ri, :ri)');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':ri': ['Hacksaw', 'Screwdriver'] });
                });
        });
    
        it('4.1.4 SET prepend element to list', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .set().attribute('ri').prepend(['Hacksaw', 'Screwdriver'])
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('SET ri = list_append(:ri, ri)');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':ri': ['Hacksaw', 'Screwdriver'] });
                });
        });
    
        it('4.1.5 SET if attribute not exist', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .set().attribute('Price').ifNotExists().equal(200)
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('SET Price = if_not_exists(Price, :Price)');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':Price': 200 });
                });
        });
    
        it('4.1.6 REMOVE', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .remove().attribute('Brand')
                .attribute('InStock')
                .attribute('QuantityOnHand')
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('REMOVE Brand, InStock, QuantityOnHand');
                });
        });
    
        it('4.1.7 ADD', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .add().attribute('QuantityOnHand').value(5)
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('ADD QuantityOnHand :QuantityOnHand');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':QuantityOnHand': 5 });
                });
        });
    
        it('4.1.8 ADD elements to a set', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .add().attribute('Color').values(['Orange', 'Purple'])
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('ADD Color :Color');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':Color': ['Orange', 'Purple'] });
                });
        });
    
        it('4.1.9 DELETE elements from a set', () => {
            let db = new SimpleDB({ dummy: 'dummy' });
            return db.update()
                .tableName('ProductCatalog')
                .primaryKey('Id').equal(123)
                .delete().attribute('Color').values(['Yellow', 'Purple'])
                .dryRun()
                .then(params => {
                    expect(params.TableName).to.be.equal('ProductCatalog');
                    expect(params.Key).to.deep.equal({ Id: 123 });
                    expect(params.UpdateExpression).to.be.equal('DELETE Color :Color');
                    expect(params.ExpressionAttributeValues).to.deep.equal({ ':Color': ['Yellow', 'Purple'] });
                });
        });
    });

    describe('4.2 Query Test: SET', () => {
        it('4.2.1 SET', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('username').equal('alice')
                .set().attribute('age').equal(20)
                .set().attribute('job').equal('retired')
                .returnValues().allNew()
                .run()
                .then(result => {
                    let attributes = result.Attributes;
                    expect(attributes.age).to.equal(20);
                    expect(attributes.job).to.equal('retired');
                });
        });

        it('4.2.2 SET with condition (reject case)', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('username').equal('Alice Tan')
                .where().attribute('job').equal('freelance')
                .returnValues().allNew()
                .run().should.be.rejected;
        });

        it('4.2.3 SET with condition (resolved case)', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('username').equal('Alice Tan')
                .where().attribute('job').equal('retired')
                .returnValues().allNew()
                .run()
                .then(result => {
                    let attributes = result.Attributes;
                    expect(attributes.username).to.equal('Alice Tan');
                });
        });

        it('4.2.4 SET nested map attributes', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('phone.home[0]').equal('00000000')
                .set().attribute('phone.home[1]').equal('11111111')
                .returnValues().allNew()
                .run()
                .then(result => {
                    let attributes = result.Attributes;
                    expect(attributes.phone.home).to.deep.equal([
                        '00000000',
                        '11111111',
                    ]);
                });
            
        });

        it('4.2.5 SET append element to list', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('phone.home').append(['22222222', '33333333'])
                .returnValues().allNew()
                .run()
                .then(result => {
                    let attributes = result.Attributes;
                    expect(attributes.phone.home).to.deep.equal([
                        '00000000',
                        '11111111',
                        '22222222',
                        '33333333',
                    ]);
                });
        });

        it('4.2.6 SET prepend element to list', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('phone.home').prepend(['10000001', '10000000'])
                .returnValues().allNew()
                .run()
                .then(result => {
                    let attributes = result.Attributes;
                    expect(attributes.phone.home).to.deep.equal([
                        '10000001',
                        '10000000',
                        '00000000',
                        '11111111',
                        '22222222',
                        '33333333',
                    ]);
                });
        });

        it('4.2.7 SET if attribute not exist (resolved case)', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('country').ifNotExists().equal('Singapore')
                .returnValues().allNew()
                .run()
                .then(item => {
                    let attributes = item.Attributes;
                    expect(attributes.country).to.be.equal('Singapore');
                });
        })
        
        it('4.2.8 SET if attribute not exist (reject case)', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('123')
                .set().attribute('country').ifNotExists().equal('Malaysia')
                .returnValues().allNew()
                .run()
                .then(item => {
                    let attributes = item.Attributes;
                    expect(attributes.country).to.be.equal('Singapore');
                });
        })
    });

    describe('4.3 Query Test: REMOVE', () => { 
        beforeEach(() => {
            AWS.config.update(config);
            let dynamodb = new AWS.DynamoDB();
            let docClient = new AWS.DynamoDB.DocumentClient({
                options: dynamodb,
            });
            return new Promise((resolve, reject) => {
                docClient.put({
                    TableName: TABLE_NAME,
                    Item: {
                        userId: '456',
                        timestamp: timestampValue,
                        job: 'retired',
                        age: 25
                    }
                }, (err, data) => {
                    if (err) {
                        reject(err);
                        return;
                    }

                    resolve(data);
                })
            })
        });

        it('4.3.1 REMOVE', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('456')
                .remove().attribute('job').attribute('age')
                .returnValues().allNew()
                .run()
                .then(item => {
                    expect(item).to.not.have.property('job');
                    expect(item).to.not.have.property('age');
                });
        });
       
    });

    describe('4.4 Query Test: ADD', () => { 
        beforeEach(() => {
            AWS.config.update(config);
            let dynamodb = new AWS.DynamoDB();
            let docClient = new AWS.DynamoDB.DocumentClient({
                options: dynamodb,
            });
            return new Promise((resolve, reject) => {
                docClient.put({
                    TableName: TABLE_NAME,
                    Item: {
                        userId: '333',
                        timestamp: timestampValue,
                        job: 'retired',
                        age: 25,
                    }
                }, (err, data) => {
                    if (err) {
                        reject(err);
                        return;
                    }

                    resolve(data);
                })
            })
        });

        it('4.4.1 ADD', () => {
            let db = new SimpleDB(config);
            return db.update()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal('333')
                .add().attribute('age').value(1)
                .returnValues().allNew()
                .run()
                .then(result => {
                    let attributes = result.Attributes;
                    expect(attributes.age).to.be.equal(26);
                })
        });

        // TODO: currently didnt work
        it('4.4.2 ADD elements to a set -- CURRENTLY NOT WORKING', () => {
            // let db = new SimpleDB(config);
            // return db.update()
            //     .tableName(TABLE_NAME)
            //     .primaryKey('userId').equal('333')
            //     .add().attribute('countries').values(['Indonesia'])
            //     .returnValues().allNew()
            //     .run().then(r => console.log(r) )
            //     .then(result => {
            //         console.log(result);
            //         let attributes = result.Attributes;
            //         expect(attributes.countries).to.deep.equal([
            //             'Indonesia'
            //         ]);
            //     })
        })

        // TODO: currently didnt work

        it('4.4.3 DELETE elements to a set -- CURRENTLY NOT WORKING', () => {
            // let db = new SimpleDB(config);
            // return db.update()
            //     .tableName(TABLE_NAME)
            //     .primaryKey('userId').equal('333')
            //     .delete().attribute('countries').values(['Indonesia'])
            //     .returnValues().allNew()
            //     .run().then(r => console.log(r) )
            //     .then(result => {
            //         console.log(result);
            //         let attributes = result.Attributes;
            //         expect(attributes.countries).to.deep.equal([
            //             'Indonesia'
            //         ]);
            //     })
        })


        

    });
    
});