const SimpleDB = require('../dynamodb');
const chai = require('chai');
const chaiAsPromised = require('chai-as-promised');
chai.use(chaiAsPromised);
const expect = chai.expect;
chai.should();

const AWS = require('aws-sdk');

const config = {
    region: 'local',
    endpoint: 'http://127.0.0.1:8000',
    accessKeyId: 'not-important',
    secretAccessKey: 'no-important'
}

const TABLE_NAME ='MyQueryTestTable'

let params = {
    TableName: TABLE_NAME,
    KeySchema: [
        {
            AttributeName: 'userId',
            KeyType: 'HASH'
        },

        {
            AttributeName: 'ts',
            KeyType: 'RANGE'
        },

    ],
    AttributeDefinitions: [
        {
            AttributeName: 'userId',
            AttributeType: 'S'
        },

        {
            AttributeName: 'ts',
            AttributeType: 'S'
        },

    ],
    ProvisionedThroughput:  {
        ReadCapacityUnits: 1,
        WriteCapacityUnits: 1
    }
}

let timestampValue = '2018-04-26T11:51:14.299Z';
let timestampValue2 = '2018-04-26T11:51:15.299Z';
let timestampValue3 = '2018-04-26T11:51:16.299Z';
let timestampValue4 = '2018-04-26T11:51:17.299Z';
let timestampValue5 = '2018-04-26T11:51:18.299Z';


function createTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.createTable(params, function (err, data) {
            if (err) {
                reject(err);
                return;
            }
 
            resolve(data);
        });
    })
}

function deleteTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.deleteTable({ TableName: TABLE_NAME }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve(data);
        })
    })
}

function configureDocClient() {
    AWS.config.update(config);
    let dynamodb = new AWS.DynamoDB();
    let docClient = new AWS.DynamoDB.DocumentClient({
        options: dynamodb,
    });

    return docClient;
}

function prepopulateTable(docClient, item) {
    return new Promise((resolve, reject) => {
        docClient.put({
            TableName: TABLE_NAME,
            Item: item
        }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve();
        })
    })
}

function populateFewItems(docClient) {
    let p1 = prepopulateTable(docClient,{ userId: '123', ts: timestampValue });
    let p2 = prepopulateTable(docClient,{ userId: '123', ts: timestampValue2 });
    let p3 = prepopulateTable(docClient,{ userId: '123', ts: timestampValue3 });
    let p4 = prepopulateTable(docClient,{ userId: '123', ts: timestampValue4 });
    let p5 = prepopulateTable(docClient,{ userId: '123', ts: timestampValue5 });

    return Promise.all([p1, p2, p3, p4, p5]);
}

before(() => {
    let docClient = configureDocClient();

    return createTable().then(() => {
        return populateFewItems(docClient);
    })
        .catch(err => {
            configureDocClient();
            return populateFewItems(docClient);
        });
});

after(() => {
    return deleteTable();
});

describe('5 DynamoDB Query Operation', () => {
    describe('5.1 Param Test', () => {
        describe('5.1.1 Primary Key', () => {
            it('5.1.1.1 Primary Key', () => {
                let db = new SimpleDB({ dummy: 'dummy' });
                return db.query()
                    .tableName('MyTable')
                    .primaryKey('userId', 'abc')
                    .dryRun()
                    .then(params => {
                        expect(params.TableName).to.be.equal('MyTable');
                        expect(params.KeyConditionExpression).to.be.equal('userId = :userId');
                        expect(params.ExpressionAttributeValues).to.deep.equal({ ':userId': 'abc' });
                    });
            })

            it('5.1.1.2 Primary Key with Equal', () => {
                let db = new SimpleDB({ dummy: 'dummy' });
                return db.query()
                    .tableName('MyTable')
                    .primaryKey('userId').equal('abc')
                    .dryRun()
                    .then(params => {
                        expect(params.TableName).to.be.equal('MyTable');
                        expect(params.KeyConditionExpression).to.be.equal('userId = :userId');
                        expect(params.ExpressionAttributeValues).to.deep.equal({ ':userId': 'abc' });
                    });
            })
        });


        describe('5.1.2 Sort Key', () => {
            it('5.1.2.1 Sort Key (legacy)', () => {
                let db = new SimpleDB({ dummy: 'dummy' });
                return db.query()
                    .tableName('MyTable')
                    .primaryKey('userId').equal('abc')
                    .sortKey('createdTime', 123, '=')
                    .dryRun()
                    .then(params => {
                        expect(params.TableName).to.be.equal('MyTable');
                        expect(params.KeyConditionExpression).to.be.equal('userId = :userId AND createdTime = :createdTime');
                        expect(params.ExpressionAttributeValues).to.deep.equal({ ':userId': 'abc', ':createdTime': 123 });
                    });
            })

            it('5.1.2.2 Sort Key', () => {
                let db = new SimpleDB({ dummy: 'dummy' });
                return db.query()
                    .tableName('MyTable')
                    .primaryKey('userId').equal('abc')
                    .sortKey('createdTime').equal(123)
                    .dryRun()
                    .then(params => {
                        expect(params.TableName).to.be.equal('MyTable');
                        expect(params.KeyConditionExpression).to.be.equal('userId = :userId AND createdTime = :createdTime');
                        expect(params.ExpressionAttributeValues).to.deep.equal({ ':userId': 'abc', ':createdTime': 123 });
                    });
            }).timeout(5000);

            it('5.1.2.3 Sort query in ascending order', () => {
                let db = new SimpleDB({ dummy: 'dummy' });
                return db.query()
                    .tableName('MyTable')
                    .primaryKey('userId').equal('abc')
                    .sortKey().ascendingOrder()
                    .dryRun()
                    .then(params => {
                        expect(params.TableName).to.be.equal('MyTable');
                        expect(params.KeyConditionExpression).to.be.equal('userId = :userId');
                        expect(params.ExpressionAttributeValues).to.deep.equal({ ':userId': 'abc' });
                        expect(params.ScanIndexForward).to.be.equal(true);
                    });
            }).timeout(5000);

            it('5.1.2.4 Sort query in descending order', () => {
                let db = new SimpleDB({ dummy: 'dummy' });
                return db.query()
                    .tableName('MyTable')
                    .primaryKey('userId').equal('abc')
                    .sortKey().descendingOrder()
                    .dryRun()
                    .then(params => {
                        expect(params.TableName).to.be.equal('MyTable');
                        expect(params.KeyConditionExpression).to.be.equal('userId = :userId');
                        expect(params.ExpressionAttributeValues).to.deep.equal({ ':userId': 'abc' });
                        expect(params.ScanIndexForward).to.be.equal(false);
                    });
            }).timeout(5000);
        });
    });

    describe('5.2 Query Test', () => {
        describe('5.2.1 Primary Key Test', () => {
            it('5.2.1.1 Primary Key', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId', '123')
                    .run()
                    .then(result => {
                        let items = result.Items;
                        expect(items[0]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue
                        });
                        expect(items[1]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue2
                        });
                    });
            });

            it('5.2.1.2 Primary Key with Equal', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId').equal('123')
                    .run()
                    .then(result => {
                        let items = result.Items;
                        expect(items[0]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue
                        });
                        expect(items[1]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue2
                        });
                    });
            })
        });    

        describe('5.2.2 Sort Key Test', () => {
            it('5.2.2.1 Sort Key (legacy)', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId').equal('123')
                    .sortKey('ts', timestampValue, '=')
                    .run()
                    .then(result => {   
                        let items = result.Items;
                        expect(items[0]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue
                        });
                    });
            });

            it('5.2.2.2 Sort Key', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId').equal('123')
                    .sortKey('ts').equal(timestampValue)
                    .run()
                    .then(result => {   
                        let items = result.Items;
                        expect(items[0]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue
                        });
                    });
            });

            it('5.2.2.3 Sort query in ascending order', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId').equal('123')
                    .sortKey().ascendingOrder()
                    .run()
                    .then(result => {   
                        let items = result.Items;
                        expect(items[0]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue
                        });
                        expect(items[1]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue2
                        });
                    });
            });

            it('5.2.2.4 Sort query in descending order', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId').equal('123')
                    .sortKey().descendingOrder()
                    .run()
                    .then(result => {   
                        let items = result.Items;
                        expect(items[0]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue5
                        });
                        expect(items[1]).to.deep.equal({
                            userId: '123',
                            ts: timestampValue4
                        });
                        
                    });
            });
        });

        describe('5.2.3 Paging Test', () => {
            it('5.2.3.1 Limit Test', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId').equal('123')
                    .limit(3)
                    .run()
                    .then(result => {
                        let items = result.Items;
                        expect(items.length).to.be.equal(3);               
                    });
            })

            it('5.2.3.1 Next Page Test', () => {
                let db = new SimpleDB(config);
                return db.query()
                    .tableName(TABLE_NAME)
                    .primaryKey('userId').equal('123')
                    .limit(3)
                    .run()
                    .then(result => {
                        let items = result.Items;
                        expect(items.length).to.be.equal(3);   
                        expect(result.LastEvaluatedKey).to.deep.equal({ userId: '123', ts: '2018-04-26T11:51:16.299Z' });
                        expect(result.hasNextPage()).to.be.true;
                        return result.nextPage()
                            .then(res => {
                                expect(res.hasNextPage()).to.be.false;
                                expect(res.Items.length).to.be.equal(2);
                            });
                    });
            })
        })
    });
});