const SimpleDB = require('../dynamodb');
const chai = require('chai');
const chaiAsPromised = require('chai-as-promised');
chai.use(chaiAsPromised);
const expect = chai.expect;
chai.should();

const AWS = require('aws-sdk');

const config = {
    region: 'local',
    endpoint: 'http://127.0.0.1:8000',
    accessKeyId: 'not-important',
    secretAccessKey: 'no-important'
}

const TABLE_NAME = 'MyCommonTestTable'

let params = {
    TableName: TABLE_NAME,
    KeySchema: [
        {
            AttributeName: 'userId',
            KeyType: 'HASH'
        },

        {
            AttributeName: 'ts',
            KeyType: 'RANGE'
        },

    ],
    AttributeDefinitions: [
        {
            AttributeName: 'userId',
            AttributeType: 'N'
        },

        {
            AttributeName: 'ts',
            AttributeType: 'S'
        },

    ],
    ProvisionedThroughput:  {
        ReadCapacityUnits: 1,
        WriteCapacityUnits: 1
    },
    GlobalSecondaryIndexes: [
        {
            IndexName: 'MyGlobalSecondaryIndex',
            KeySchema: [
                {
                    AttributeName: 'ts',
                    KeyType: 'HASH'
                }
            ],
            Projection: {
                ProjectionType: "ALL"
            },
            ProvisionedThroughput:  {
                ReadCapacityUnits: 1,
                WriteCapacityUnits: 1
            },
        }
    ],
    // LocalSecondaryIndexes: [
    //     {
    //         IndexName: 'MyLocalSecondaryIndex',
    //         KeySchema: [
    //             {
    //                 AttributeName: 'ts',
    //                 KeyType: 'HASH'
    //             }
    //         ],
    //         Projection: {
    //             ProjectionType: "ALL"
    //         },
    //         ProvisionedThroughput:  {
    //             ReadCapacityUnits: 1,
    //             WriteCapacityUnits: 1
    //         },
    //     }
    // ]
}

let timestampValue = '2018-04-26T11:51:14.299Z';
let timestampValue2 = '2018-04-26T11:51:15.299Z';
let timestampValue3 = '2018-04-26T11:51:16.299Z';
let timestampValue4 = '2018-04-26T11:51:17.299Z';
let timestampValue5 = '2018-04-26T11:51:18.299Z';
let timestampValue6 = '2018-04-26T11:51:19.299Z';
let timestampValue7 = '2018-04-26T11:51:20.299Z';
let timestampValue8 = '2018-04-26T11:51:21.299Z';


function createTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.createTable(params, function (err, data) {
            if (err) {
                reject(err);
                return;
            }
 
            resolve(data);
        });
    })
}

function deleteTable() {
    let dynamodb = new AWS.DynamoDB(config);

    return new Promise((resolve, reject) => {
        dynamodb.deleteTable({ TableName: TABLE_NAME }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve(data);
        })
    })
}

function configureDocClient() {
    AWS.config.update(config);
    let dynamodb = new AWS.DynamoDB();
    let docClient = new AWS.DynamoDB.DocumentClient({
        options: dynamodb,
    });

    return docClient;
}

function prepopulateTable(docClient, item) {
    return new Promise((resolve, reject) => {
        docClient.put({
            TableName: TABLE_NAME,
            Item: item
        }, (err, data) => {
            if (err) {
                reject(err);
                return;
            }

            resolve();
        })
    })
}

function populateFewItems(docClient) {
    let p1 = prepopulateTable(docClient,{ userId: 123, ts: timestampValue });
    let p2 = prepopulateTable(docClient,{ userId: 123, ts: timestampValue2 });
    let p3 = prepopulateTable(docClient,{ userId: 123, ts: timestampValue3 });
    let p4 = prepopulateTable(docClient,{ userId: 124, ts: timestampValue4 });
    let p5 = prepopulateTable(docClient,{ userId: 124, ts: timestampValue5 });
    let p6 = prepopulateTable(docClient, { userId: 125, ts: timestampValue6, age: 25 });
    let p7 = prepopulateTable(docClient, { userId: 126, ts: timestampValue7 });
    let p8 = prepopulateTable(docClient, { userId: 127, ts: timestampValue8, age: 45 });
    
    return Promise.all([p1, p2, p3, p4, p5]);
}

before(() => {
    let docClient = configureDocClient();

    return createTable().then(() => {
        return populateFewItems(docClient);
    })
        .catch(err => {
            configureDocClient();
            return populateFewItems(docClient);
        });
});

after(() => {
    return deleteTable();
});

describe('7 DynamodDB Common Operation', () => {
    describe('7.1 Attribute Comparison', () => { 
        it('7.1.1 Attribute Equal', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').equal(123)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(3);
                    expect(result.hasNextPage()).to.be.false;
                });
        }).timeout(5000);
    
        it('7.1.2 Attribute NotEqual', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').notEqual(123)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(5);
                    expect(result.hasNextPage()).to.be.false;
                }); 
        }).timeout(5000);
    
        it('7.1.3 Attribute LessThan', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').lessThan(124)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(3);
                    expect(result.hasNextPage()).to.be.false;
                });
        }).timeout(5000);
    
        it('7.1.4 Attribute MoreThan', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').greaterThan(123)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(5);
                    expect(result.hasNextPage()).to.be.false;
                }); 
        }).timeout(5000);
    
        it('7.1.5 Attribute LessThanOrEqual', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').lessThanOrEqual(124)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(5);
                    expect(result.hasNextPage()).to.be.false;
                }); 
        }).timeout(5000);
    
        it('7.1.6 Attribute MoreThanOrEqual', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').greaterThanOrEqual(124)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(5);
                    expect(result.hasNextPage()).to.be.false;
                }); 
        }).timeout(5000);
    });
    
    describe('7.2 Attribute Range', () => {
        it('7.2.1 Attribute Between', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').between(124, 126) // inclusive
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(4);
                    expect(result.hasNextPage()).to.be.false;
                });
        }).timeout(5000);

        it('7.2.2 Attribute In', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').in([124,127])
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(3);
                    expect(result.hasNextPage()).to.be.false;
                });
        }).timeout(5000);
    });    

    describe('7.3 Attribute Functions', () => {
        it('7.3.1 Attribute Exists', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('age').exists()
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(2);
                });
        }).timeout(5000);

        it('7.3.2 Attribute Not Exists', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('age').notExists()
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(6);
                });
        }).timeout(5000);

        it('7.3.3 Attribute Begins With', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('ts').beginsWith('2018-04-26T11:51:1')
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(6);
                });
        }).timeout(5000);

        it('7.3.4 Attribute Contains', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('ts').contains('51:1')
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(6);
                });
        }).timeout(5000);

        describe('7.3.5 Attribute Size', () => {
            it('7.3.5.1 Size Equal', () => {
                let db = new SimpleDB(config);
                return db.scan()
                    .tableName(TABLE_NAME)
                    .attribute('ts').size().equal(24)
                    .run()
                    .then(result => {
                        expect(result.Count).to.be.equal(8);
                    });
            });


            it('7.3.5.2 Size Not Equal', () => {
                let db = new SimpleDB(config);
                return db.scan()
                    .tableName(TABLE_NAME)
                    .attribute('ts').size().notEqual(24)
                    .run()
                    .then(result => {
                        expect(result.Count).to.be.equal(0);
                    });
            });

            it('7.3.5.3 Size Less Than', () => {
                let db = new SimpleDB(config);
                return db.scan()
                    .tableName(TABLE_NAME)
                    .attribute('ts').size().lessThan(20)
                    .run()
                    .then(result => {
                        expect(result.Count).to.be.equal(0);
                    });
            });

            it('7.3.5.4 Size Greater Than', () => {
                let db = new SimpleDB(config);
                return db.scan()
                    .tableName(TABLE_NAME)
                    .attribute('ts').size().greaterThan(20)
                    .run()
                    .then(result => {
                        expect(result.Count).to.be.equal(8);
                    });
            });

            it('7.3.5.5 Size Less Than or Equal', () => {
                let db = new SimpleDB(config);
                return db.scan()
                    .tableName(TABLE_NAME)
                    .attribute('ts').size().lessThanOrEqual(24)
                    .run()
                    .then(result => {
                        expect(result.Count).to.be.equal(8);
                    });
            });

            it('7.3.5.6 Size Greater Than or Equal', () => {
                let db = new SimpleDB(config);
                return db.scan()
                    .tableName(TABLE_NAME)
                    .attribute('ts').size().greaterThanOrEqual(20)
                    .run()
                    .then(result => {
                        expect(result.Count).to.be.equal(8);
                    });
            });

        });
    });    

    describe('7.4 Attribute Join Condition', () => { 
        it('7.4.1 Attribute AND', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').greaterThan(120).and()
                .attribute('age').greaterThan(40)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(1);
                }); 
        }).timeout(5000);

        it('7.4.2 Attribute OR', () => {
            let db = new SimpleDB(config);
            return db.scan()
                .tableName(TABLE_NAME)
                .attribute('userId').equal(123).or()
                .attribute('age').equal(45)
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(4);
                }); 
        }).timeout(5000);
    });

    describe('7.5 Projection', () => {
        it('7.5.1 Projection as Array', () => {
            let db = new SimpleDB(config);
            return db.query()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal(125)
                .select(['ts','age'])
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(1);
                }); 
        }).timeout(5000);

        it('7.5.2 Projection as String', () => {
            let db = new SimpleDB(config);
            return db.query()
                .tableName(TABLE_NAME)
                .primaryKey('userId').equal(125)
                .select('ts,age')
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(1);
                }); 
        }).timeout(5000);
    });
    
    describe('7.6 Secondary Index', () => {
        it('7.6.1 With Consistent Read', () => {
            let db = new SimpleDB(config);
            return db.query()
                .tableName(TABLE_NAME)
                .primaryKey('ts').equal(timestampValue5)
                .indexName('MyGlobalSecondaryIndex')
                // .isConsistentRead()
                .select('userId')
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(1);
                });
        }).timeout(5000);

        it('7.6.2 With ReturnedConsumedCapacity', () => {
            let db = new SimpleDB(config);
            return db.query()
                .tableName(TABLE_NAME)
                .primaryKey('ts').equal(timestampValue5)
                .indexName('MyGlobalSecondaryIndex')
                // .isConsistentRead()
                .returnedConsumedCapacity('INDEXES')
                .select('userId, ts')
                .run()
                .then(result => {
                    expect(result.Count).to.be.equal(1);
                });
        }).timeout(5000);

    });        

  
});